#!/usr/local/bin/perl

# http://article.gmane.org/gmane.science.linguistics.wikipedia.technical/38154/match=new+statistics+stuff
# http://svn.wikimedia.org/viewvc/mediawiki/trunk/webstatscollector/
# https://bugzilla.wikimedia.org/show_bug.cgi?id=13541
# http://de.wikipedia.org/w/api.php?action=query&meta=siteinfo&siprop=general|namespaces|namespacealiases

# %20 -> underscore
# rest naar chr(%xx) waarde
# namespace -> number

# frequenty distribution hits per file per first letter _-> manifest crawler
# first letter uppercase -> sort (in sections per first two chars ?)
# discard entries < x hits ? (crawler)
# read all files at same time

  use CGI qw(:all);
  use IO::Uncompress::Gunzip qw(gunzip $GunzipError) ; # install IO-Compress-Zlib
  use IO::Compress::Gzip     qw(gzip   $GzipError) ;   # install IO-Compress-Zlib

  $| = 1; # flush screen output
  $true  = 1 ;
  $false = 0 ;

  $dir0 = "D:/Wikipedia_Visitors" ;
  open LOG, ">>", "$dir0/WikiStatsCompactVisitorstats.log" ;

  # &CompactVisitorStats ;
  &UncompactVisitorStats ; # test to see if process is revertible

  &Log ("Ready") ;
  close LOG ;
  exit ;

sub UncompactVisitorStats
{
  my $file_in = "D:/Wikipedia_Visitors/full_day/2008-07-pagecounts/pagecounts-20080701_fd.gz" ;
  my $dir_out = "D:/Wikipedia_Visitors/test" ;
  $in_gz = IO::Uncompress::Gunzip->new ($file_in) or die "IO::Uncompress::Gunzip failed for '$file_in': $GunzipError\n";
  binmode $in_gz ;

  for ($h=0 ; $h<=23 ; $h++)
  {
    $time = sprintf ("%02d",$h) . "0000" ;
    $file_out = "$dir_out/2008-07-pagecounts/pagecounts-20080701-$time.gz" ;
    $out_gz [$h] = IO::Compress::Gzip->new ($file_out) or die "IO::Compress::Gzip failed: $GzipError\n";
    binmode $out_gz [$h] ;
  }

  while ($line = <$in_gz>)
  {
    chomp ($line) ;
#   print "$line\n" ;
   if ($lines++ > 1000000) { exit ; }
    ($lang,$title,$counts) = split (" ", $line) ;
    $lang =~ s/\.z// ;
    $lang =~ s/\.y/2/ ;
    $counts =~ s/^\d+// ; # remove (redundant) preceding total
    while ($counts ne "")
    {
      $letter = substr ($counts,0,1) ;
      $counts = substr ($counts,1) ;
      ($count = $counts) =~ s/^(\d+).*$/$1/ ;
      $counts =~ s/^\d+(.*)$/$1/ ;
      $h = ord ($letter) - ord ('A') ;
      $file = $out_gz [$h] ;
      print $file "$lang $title $count\n" ;
    }

  }

  for ($h=0 ; $h<=23 ; $h++)
  {
    $out_gz [$h] -> close () ;
  }
}

sub CompactVisitorStats
{
  $dir_out1 = "full_day" ;
  $dir_out2 = "full_day_10_views" ;
  chdir ($dir0) || die "Cannot chdir to $dir0\n";
  mkdir $dir_out1 ;
  mkdir $dir_out2 ;

  # make this smarter some day
# &ProcessMonth (2008,1) ;
# &ProcessMonth (2008,2) ;
# &ProcessMonth (2008,3) ;
# &ProcessMonth (2008,4) ;
# &ProcessMonth (2008,5) ;
# &ProcessMonth (2008,6) ;
# &ProcessMonth (2008,7) ;
# &ProcessMonth (2008,8) ;
}

sub ProcessMonth
{
  my $year  = shift ;
  my $month = sprintf ("%02d", shift) ;

  $dir0 =~ s/[\\\/]$// ;

  chdir "$dir0/$dir_out1" || die "Cannot chdir to $dir0/$dir_out1\n";
  mkdir ("$year-$month-pagecounts") ;
  chdir "$dir0/$dir_out2" || die "Cannot chdir to $dir0/$dir_out2\n";
  mkdir ("$year-$month-pagecounts") ;

  $dir_out1_full = "$dir0/$dir_out1/$year-$month-pagecounts" ;
  $dir_out2_full = "$dir0/$dir_out2/$year-$month-pagecounts" ;

  $dir_in = "$dir0/$year-$month-pagecounts" ;
  &Log ("Process year $year month $month from '$dir_in'\n") ;
  chdir ($dir_in) || die "Cannot chdir to $dir_in\n";
  local (*DIR);

  # collect files for one day
  for ($d = 1 ; $d <= 31 ; $d++)
  {
    opendir (DIR, ".");
    $day = sprintf ("%02d", $d) ;
    @files = () ;
    while ($file_in = readdir (DIR))
    {
      if ($file_in !~ /^pagecounts-\d{6,6}$day-\d{6,6}.gz$/)
      { next ; }
      push @files, $file_in ;
    }

    $filecnt = $#files+1 ;
    if ($filecnt == 24)
    {
      @files = sort { substr ($a, 20,2) <=> substr ($b, 20,2)} @files ;
      &MergeFiles ($year,$month,$day,@files) ;
    }
    else
    { &Log ("$day/$month/$year: $filecnt files found -> do not merge\n") ; }

    closedir (DIR, ".");
  }
}

sub GetViewDistribution
{
  open OUT, ">", "Views.csv" ;
  foreach $file_in (@files)
  {
    ($hour = $file_in) =~ s/^pagecounts-\d+-(\d\d)\d+\.gz$/$1/ ;
    $hour = chr(ord('A')+$hour) ;
    &Log ("Process $hour $file_in\n") ;

    $in_gz1 = IO::Uncompress::Gunzip->new ($file_in) or die "IO::Uncompress::Gunzip failed: $GunzipError\n";
    while ($line = <$in_gz1>)
    {
      ($lang,$title,$count,$dummy) = split (' ', $line) ;
      if (($lang eq "en") && ($title !~ /:/)) # only en: and namespace 0
      {
        $tot {$hour} += $count ;
        if ($count < 3)
        { @counts {$hour . substr ($title,0,1)}++ ; }
      }
    }
    $in_gz1->close () ;
  }

  print OUT "," ;
  foreach $hour ('A'..'X')
  {  print OUT $hour . ", " ; }
  print OUT "\n" ;

  print OUT "," ;
  foreach $hour ('A'..'X')
  {  print OUT $tot {$hour} . ", " ; }
  print OUT "\n" ;

  for ($c=0; $c < 256; $c++)
  {
    # do not print chars " and , as such: confuses csv format
    if ($c < 33)
    { print OUT "chr($c), " ; }
    elsif (chr($c) eq '"')
    { print OUT "dquote, " ; }
    elsif (chr($c) eq ',')
    { print OUT "comma, " ; }
    else
    { print OUT chr($c) . ", " ; }

    foreach $hour ('A'..'X')
    {  print OUT (0+@counts {$hour.chr($c)}) , ", " ; }

    if ($c < 255)
    { print OUT "\n" ; }
  }
  close OUT ;
}

sub MergeFiles
{
  $time_start = time ;
  $lines = 0 ;

  my $year  = shift ;
  my $month = shift ;
  my $day   = shift ;
  my @files = @_ ;
  &Log ("$day/$month/$year: merge files\n") ;

  undef @in_gz ;
  undef $file_open ;
  my $time_start = time ;

  # $file_out = "pagecounts-$year$month$day_full_day" ;
  # open OUT, ">", $file_out ;
  # binmode $file_out ;

  my $file_out1 = "$dir_out1_full/pagecounts-$year$month$day" . "_fd.gz" ;
  my $file_out2 = "$dir_out2_full/pagecounts-$year$month$day" . "_fd10.gz" ;
  # print "File_out1 $file_out1\n" ;
  # print "File_out2 $file_out2\n" ;

  my $out_gz1 = IO::Compress::Gzip->new ($file_out1) or die "IO::Compress::Gzip failed: $GzipError\n";
  my $out_gz2 = IO::Compress::Gzip->new ($file_out2) or die "IO::Compress::Gzip failed: $GzipError\n";
  binmode $out_gz1 ;
  binmode $out_gz2 ;

  $files_in_open = 0 ;

  foreach $file_in (@files)
  {
#    if ($files++ == 0)
#    {
#      ($day  = $file_in) =~ s/^pagecounts-(\d+)-\d\d\d+\.gz$/$1/ ;
#      print "Process $day '$file_in'\n" ;
#      $file_out = "pagecounts-$day.gz" ;
#    }

    ($hour = $file_in) =~ s/^pagecounts-\d+-(\d\d)\d+\.gz$/$1/ ;
    $hour = (0+$hour) ;
    # print "            file found '$file_in'\n" ;

    $in_gz [$hour] = IO::Uncompress::Gunzip->new ($file_in) or die "IO::Uncompress::Gunzip failed for '$file_in': $GunzipError\n";
    binmode $in_gz [$hour] ;
    $files_in_open++ ;
    $file_in_open [$hour] = $true ;
    $file = $in_gz [$hour] ;
    $line = <$file> ;
    $line =~ s/^(\w+)2 /$1.y /o  ;
    $line =~ s/^(\w+) /$1.z /o  ;

    ($lang,$title,$count [$hour],$dummy) = split (' ', $line) ;
    $key [$hour] = "$lang $title" ;
  }

  $key_low_prev = "" ;
  while ($files_in_open > 0)
  {
    $key_low = "\xFF";
    for ($hour = 0 ; $hour < 24 ; $hour++)
    {
      if (($files_in_open == 24) || ($file_in_open [$hour]))
      {
        if ($key [$hour] lt $key_low)
        { $key_low = $key [$hour] ; }
      }
    }
    $counts = "" ;
    $total  = 0 ;
    for ($hour = 0 ; $hour < 24 ; $hour++)
    {
      if (($files_in_open == 24) || ($file_in_open [$hour]))
      {
        if ($key [$hour] eq $key_low)
        {
          $counts .= chr ($hour+ord('A')) . $count [$hour] ;
          $total += $count [$hour] ;
          $file = $in_gz [$hour] ;
          # $line = <$file> ;
          if ($line = <$file>) #  =~ /^a/)
          {
            $line =~ s/^([\w\-]+)2 /$1.y /o  ;
            $line =~ s/^([\w\-]+) /$1.z /o  ;
           ($lang,$title,$count [$hour],$dummy) = split (' ', $line) ;
            $key [$hour] = "$lang $title" ;
          }
          else
          {
            $in_gz [$hour] -> close () ;
            $files_in_open-- ;
            $file_in_open [$hour] = $false ;
            $key [$hour] = "\xFF";
          }
        }
      }
    }
    if ($lines == 0) { &Log ("\n[lines done, files open] key\n") ; }
    if (++$lines % 1000000 == 0) { &Log ("[$lines, $files_in_open] $key_low\n") ; }
    if ($key_low_prev ge $key_low) { die "Sequence error: '$key_low_prev' ge '$key_low'\n" ; }

    # print OUT "$key_low $total$counts\n" ;
    print $out_gz1 "$key_low $total$counts\n" ;
    if ($total >= 10)
    { print $out_gz2 "$key_low $total$counts\n" ; }

    $key_low_prev = $key_low ;
  # print "OUT $key_low $counts\n" ;
  }

  &Log ("[$lines, $files_in_open] $key_low\n") ;
# close OUT ;

  $out_gz1->close() ;
  $out_gz2->close() ;

  &Log ("Processed in " . (time-$time_start) . " seconds\n\n") ;
}

sub Compact
{
  my $day = shift ;
  &Log ("Compact files for $day\n") ;

  $file_in  = "pagecounts-$day.out" ;
  $file_out1 = "pagecounts-${day}_all.gz" ;
  $file_out2 = "pagecounts-${day}_10plus.gz" ;
  open IN, "<", $file_in ;
  binmode $file_in ;

  my $out_gz1 = IO::Compress::Gzip->new ($file_out1) or die "IO::Compress::Gzip failed: $GzipError\n";
  my $out_gz2 = IO::Compress::Gzip->new ($file_out2) or die "IO::Compress::Gzip failed: $GzipError\n";

  open OUT, ">", $file_out ;
  binmode $file_out ;

  $lang_prev = "" ;
  while ($line = <IN>)
  {
    chomp ($line) ;
    ($lang, $title, $counts) = split (' ', $line) ;
    $title2 = $title ;
    $title =~ s/\%20/_/g ;
    $title =~ s/\%([0-9A-F]{2})/chr(hex($1))/ge ;
  # if ($title =~ /[\x00-\x1F]/)
  # { &Log ("> '$title2'\n") ; }
    $title =~ s/\x00-\x1F/"%" . sprintf ("%X", ord($1)) ;/ge ;
    print $out_gz1 "$lang $title $counts\n" ;
    ($counts2 = $counts) =~ s/^(\d+).*/$1/ ;
    if ($counts2 >= 10)
    { print $out_gz2 "$lang $title $counts\n" ; }
    $lang_prev = $lang ;
  }

  close IN ;
  $out_gz1->close() ;
  $out_gz2->close() ;
}


sub Log
{
  $msg = shift ;
  print $msg ;
  print LOG $msg ;
}


